import pytest
import numpy
from numpy.testing import assert_equal
from spacy_transformers.data_classes import WordpieceBatch


@pytest.fixture
def wordpieces():
    strings = [["some", "random", "strings"], ["are"], ["added", "here"]]
    shape = (len(strings), max(len(seq) for seq in strings))
    wordpieces = WordpieceBatch(
        strings=strings,
        input_ids=numpy.zeros(shape, dtype="i"),
        token_type_ids=numpy.zeros(shape, dtype="i"),
        attention_mask=numpy.zeros((shape[0], shape[1]), dtype="bool"),
        lengths=[len(seq) for seq in strings],
    )
    return wordpieces


def test_wordpieces_IO(wordpieces):
    wp_dict = wordpieces.to_dict()
    wordpieces_2 = WordpieceBatch.empty().from_dict(wp_dict)
    for key, value in wordpieces_2.to_dict().items():
        assert_equal(value, wp_dict[key])



from typing import Tuple, List, Union, TypeVar
import numpy
from thinc.types import Ragged, Ints2d, Floats2d
from .data_classes import WordpieceBatch

ArrayT = TypeVar("ArrayT", bound=Union[Ints2d, Floats2d])


def truncate_oversize_splits(
    wordpieces: WordpieceBatch, align: Ragged, max_length: int
) -> Tuple[WordpieceBatch, Ragged]:
    """Drop wordpieces from inputs that are too long. This can happen because
    the splitter is based on linguistic tokens, and the number of wordpieces
    that each token is split into is unpredictable, so we can end up with splits
    that have more wordpieces than the model's maximum.

    To solve this, we calculate a score for each wordpiece in the split,
    and drop the wordpieces with the highest scores. I can think of a few
    scoring schemes we could use:

    a) Drop the ends of longest wordpieces. This scoring would be just:
        position_in_token
    b) Drop the middles of longest wordpieces. The score would be:
        min(length - position_in_token, position_in_token)
    c) Drop all wordpieces from longest tokens. This would be:
        length
    d) Drop wordpieces from the end of the split. This would be:
        position_in_split

    The advantage of a) and b) is that they give some representation to each
    token. The advantage of c) is that it leaves a higher % of tokens with intact
    representations. The advantage of d) is that it leaves contiguous chunks of
    wordpieces intact, and drops from the end.

    I find b) most appealing, but it's also the most complicated. Let's just do
    d) for now.
    """
    if wordpieces.input_ids.shape[1] < max_length:
        return wordpieces, align
    mask = _get_truncation_mask_drop_from_end(
        wordpieces.input_ids.shape, wordpieces.lengths, align, max_length
    )
    return _truncate_tokens(wordpieces, mask), _truncate_alignment(align, mask)


def _get_truncation_mask_drop_from_end(
    shape: Tuple[int, int], split_lengths: List[int], align: Ragged, max_length: int
) -> numpy.ndarray:
    """Return a two-dimensional boolean mask, indicating whether wordpieces
    are dropped from their sequences.

    Drop wordpieces from the end of the sequence.
    """
    mask = numpy.ones(shape, dtype="bool")
    mask[:, max_length:] = 0
    return mask


def _truncate_tokens(wordpieces: WordpieceBatch, mask: numpy.ndarray) -> WordpieceBatch:
    n_seq = len(wordpieces)
    mask1d = mask.ravel()
    mask = mask.reshape((n_seq, -1))

    strings: List[List[str]] = []
    for i, seq in enumerate(wordpieces.strings):
        strings.append([])
        for j, token in enumerate(seq):
            if mask[i, j]:
                strings[-1].append(token)

    def filter_array(data: ArrayT) -> ArrayT:
        data1d = data.reshape((-1,))
        return data1d[mask1d].reshape((n_seq, -1))  # type: ignore

    filtered_token_type_ids = None
    if wordpieces.token_type_ids is not None:
        filtered_token_type_ids = filter_array(wordpieces.token_type_ids)

    return WordpieceBatch(
        strings=strings,
        input_ids=filter_array(wordpieces.input_ids),
        attention_mask=filter_array(wordpieces.attention_mask),
        lengths=[len(seq) for seq in strings],
        token_type_ids=filtered_token_type_ids,
    )


def _truncate_alignment(align: Ragged, mask: numpy.ndarray) -> Ragged:
    # We're going to have fewer wordpieces in the new array, so all of our
    # wordpiece indices in the alignment table will be off --- they'll point
    # to the wrong row. So we need to do three things here:
    #
    # 1) Adjust all the indices in align.dataXd to account for the dropped data
    # 2) Remove the dropped indices from the align.dataXd
    # 3) Calculate new align.lengths
    #
    # The wordpiece mapping is easily calculated by the cumulative sum of the
    # mask table.
    # Let's say we have [True, False, False, True]. The mapping of the dropped
    # wordpieces doesn't matter, because we can filter it with the mask. So we
    # have [0, 0, 0, 1], i.e the wordpiece that was
    # at 0 is still at 0, and the wordpiece that was at 3 is now at 1.
    mask = mask.ravel()
    idx_map = mask.cumsum() - 1
    idx_map[~mask] = -1
    # Step 1: Adjust all the indices in align.dataXd.
    new_align = idx_map[align.data.ravel()]
    # Step 2: Remove the dropped indices
    new_align = new_align[new_align >= 0]
    # Step 3: Calculate new align.lengths
    new_lengths = align.lengths.copy()
    for i in range(len(align.lengths)):
        drops = ~mask[align[i].data.ravel()]
        new_lengths[i] -= drops.sum()
    return Ragged(new_align, new_lengths)



# cython: infer_types=True, boundscheck=False
from typing import cast, Dict, List, Tuple, Callable, Set, Optional
import numpy
from spacy_alignments.tokenizations import get_alignments
from spacy.tokens import Span, Token
from thinc.api import Ops
from thinc.types import Ragged, Floats2d, Ints1d, Ints2d

from cython.operator cimport dereference as deref
from cython.operator cimport preincrement as preinc
from libc.stdint cimport uint32_t, int32_t, int64_t
from libc.stdlib cimport free
from libcpp.unordered_set cimport unordered_set
from libcpp.vector cimport vector

ctypedef unordered_set[uint32_t]* unordered_set_uint32_t_ptr


def apply_alignment(ops: Ops, align: Ragged, X: Floats2d) -> Tuple[Ragged, Callable]:
    """Align wordpiece data (X) to match tokens, and provide a callback to
    reverse it.

    This function returns a Ragged array, which represents the fact that one
    token may be aligned against multiple wordpieces. It's a nested list,
    concatenated with a lengths array to indicate the nested structure.

    The alignment is also a Ragged array, where the lengths indicate how many
    wordpieces each token is aligned against. The output ragged therefore has
    the same lengths as the alignment ragged, which means the output data
    also has the same number of data rows as the alignment. The size of the
    lengths array indicates the number of tokens in the batch.

    The actual alignment is a simple indexing operation:

        for i, index in enumerate(align.data):
            Y[i] = X[index]

    Which is vectorized via numpy advanced indexing:

        Y = X[align.data]

    The inverse operation, for the backward pass, uses the 'scatter_add' op
    because one wordpiece may be aligned against multiple tokens. So we need:

        for i, index in enumerate(align.data):
            X[index] += Y[i]

    The addition wouldn't occur if we simply did `X[index] = Y`, so we use
    the scatter_add op.
    """
    if not align.lengths.sum():
        return _apply_empty_alignment(ops, align, X)
    shape = X.shape
    indices = cast(Ints1d, align.dataXd)
    Y = Ragged(X[indices], cast(Ints1d, ops.asarray(align.lengths)))

    def backprop_apply_alignment(dY: Ragged) -> Floats2d:
        assert dY.data.shape[0] == indices.shape[0]
        dX = ops.alloc2f(*shape)
        ops.scatter_add(dX, indices, cast(Floats2d, dY.dataXd))
        return dX

    return Y, backprop_apply_alignment


def _apply_empty_alignment(ops, align, X):
    shape = X.shape
    Y = Ragged(
        ops.alloc2f(align.lengths.shape[0], X.shape[1]),
        ops.alloc1i(align.lengths.shape[0]) + 1,
    )

    def backprop_null_alignment(dY: Ragged) -> Floats2d:
        return ops.alloc2f(*shape)

    return Y, backprop_null_alignment


def get_token_positions(spans: List[Span]) -> Dict[Token, int]:
    token_positions: Dict[Token, int] = {}
    seen_docs = set()
    for span in spans:
        if span.doc in seen_docs:
            continue
        seen_docs.add(span.doc)
        for token in span.doc:
            if token not in token_positions:
                token_positions[token] = len(token_positions)
    return token_positions


def get_alignment_via_offset_mapping(
    spans: List[Span],
    offset_mapping: Ints2d,
) -> Ragged:
    if len(spans) != len(offset_mapping):
        raise ValueError("Cannot align batches of different sizes.")
    # Tokens can occur more than once, and we need the alignment of each token
    # to its place in the concatenated wordpieces array.
    token_positions = get_token_positions(spans)
    alignment: List[Set[int]] = [set() for _ in range(len(token_positions))]
    wp_start = 0
    for i, span in enumerate(spans):
        span_offset_mapping = offset_mapping[i]
        span2wp = get_span2wp_from_offset_mapping(span, span_offset_mapping)
        for token, wp_js in zip(span, span2wp):
            position = token_positions[token]
            alignment[position].update(wp_start + j for j in wp_js)
        wp_start += span_offset_mapping.shape[0]
    lengths: List[int] = []
    flat: List[int] = []
    for a in alignment:
        lengths.append(len(a))
        flat.extend(sorted(a))
    align = Ragged(
        cast(Ints1d, numpy.array(flat, dtype="i")),
        cast(Ints1d, numpy.array(lengths, dtype="i")),
    )
    return align


def get_alignment(
    spans: List[Span],
    wordpieces: List[List[str]],
    special_tokens: Optional[List[str]] = None,
) -> Ragged:
    """Compute a ragged alignment array that records, for each unique token in
    `spans`, the corresponding indices in the flattened `wordpieces` array.
    For instance, imagine you have two overlapping spans:

        [[I, like, walking], [walking, outdoors]]

    And their wordpieces are:

        [[I, like, walk, ing], [walk, ing, out, doors]]

    We want to align "walking" against [walk, ing, walk, ing], which have
    indices [2, 3, 4, 5] once the nested wordpieces list is flattened.

    The nested alignment list would be:

    [[0], [1], [2, 3, 4, 5], [6, 7]]
      I   like    walking    outdoors

    Which gets flattened into the ragged array:

    [0, 1, 2, 3, 4, 5, 6, 7]
    [1, 1, 4, 2]

    The ragged format allows the aligned data to be computed via:

    tokens = Ragged(wp_tensor[align.data], align.lengths)

    This produces a ragged format, indicating which tokens need to be collapsed
    to make the aligned array. The reduction is deferred for a later step, so
    the user can configure it. The indexing is especially efficient in trivial
    cases like this where the indexing array is completely continuous.
    """
    if len(spans) != len(wordpieces):
        raise ValueError("Cannot align batches of different sizes.")
    if special_tokens is None:
        special_tokens = []
    # Tokens can occur more than once, and we need the alignment of each token
    # to its place in the concatenated wordpieces array.
    token_positions = get_token_positions(spans)
    alignment: List[Set[int]] = [set() for _ in range(len(token_positions))]
    wp_start = 0
    for i, (span, wp_toks) in enumerate(zip(spans, wordpieces)):
        sp_toks = [token.text for token in span]
        wp_toks_filtered = wp_toks
        # In the case that the special tokens do not appear in the text, filter
        # them out for alignment purposes so that special tokens like "<s>" are
        # not aligned to the character "s" in the text. (If the special tokens
        # appear in the text, it's not possible to distinguish them from the
        # added special tokens, so they may be aligned incorrectly.)
        if not any([special in span.text for special in special_tokens]):
            wp_toks_filtered = [
                tok if tok not in special_tokens else "" for tok in wp_toks
            ]
        span2wp, wp2span = get_alignments(sp_toks, wp_toks_filtered)
        for token, wp_js in zip(span, span2wp):
            position = token_positions[token]
            alignment[position].update(wp_start + j for j in wp_js)
        wp_start += len(wp_toks)
    lengths: List[int] = []
    flat: List[int] = []
    for a in alignment:
        lengths.append(len(a))
        flat.extend(sorted(a))
    align = Ragged(
        cast(Ints1d, numpy.array(flat, dtype="i")),
        cast(Ints1d, numpy.array(lengths, dtype="i")),
    )
    return align


def get_span2wp_from_offset_mapping(span, wp_char_offsets):
    # create a mapping of char indices to spacy token indices
    cdef int span_idx = span[0].idx
    cdef int span_i = span[0].i
    cdef int char_idx, rel_token_i
    # size is +1 so we don't have to check whether the text has a trailing space
    char_to_sp_token = numpy.full((len(span.text) + 1,), -1, dtype="int32")
    for token in span:
        rel_token_i = token.i - span_i
        for char_idx in range(
                token.idx - span_idx,
                token.idx - span_idx + len(token) + 1,
        ):
            char_to_sp_token[char_idx] = rel_token_i

    # align all wordpiece tokens to one or more spacy token indices
    cdef vector[unordered_set_uint32_t_ptr] alignment
    for _ in range(len(span)):
        alignment.push_back(new unordered_set[uint32_t]())
    _get_span2wp_alignment(
        &alignment,
        numpy.ascontiguousarray(char_to_sp_token),
        char_to_sp_token.size,
        numpy.ascontiguousarray(wp_char_offsets, dtype="int64"),
        wp_char_offsets.shape[0],
    )

    # convert the alignment into a list of aligned wordpiece indices per spacy
    # token index (unsorted at this point)
    cdef unordered_set_uint32_t_ptr s
    cdef vector[unordered_set_uint32_t_ptr].iterator it_v = alignment.begin()
    cdef unordered_set[uint32_t].iterator it_s
    result: List[List[int]] = []
    while it_v != alignment.end():
        result.append([])
        s = deref(it_v)
        it_s = s.begin()
        while it_s != s.end():
            result[-1].append(deref(it_s))
            preinc(it_s)
        del s
        preinc(it_v)
    return result


cdef int _get_span2wp_alignment(
        vector[unordered_set_uint32_t_ptr]* alignment,
        int32_t[::1] char_to_sp_token,
        int char_to_sp_token_length,
        int64_t[:, ::1] wp_char_offsets,
        int wp_char_offsets_length,
    ) nogil:
    cdef int char_idx, start_idx, end_idx, token_i
    cdef int wp_j = 0
    cdef int alignment_size = alignment.size()
    while wp_j < wp_char_offsets_length:
        start_idx = wp_char_offsets[wp_j][0]
        end_idx = wp_char_offsets[wp_j][1]
        char_idx = start_idx
        while char_idx < end_idx:
            if 0 <= char_idx < char_to_sp_token_length:
                token_i = char_to_sp_token[char_idx]
            else:
                token_i = -1
            if 0 <= token_i < alignment_size:
                deref(alignment.at(token_i)).insert(wp_j)
            char_idx += 1
        wp_j += 1
    return 0



from typing import List, Tuple, Callable, Union, Dict
import copy
from pathlib import Path
from transformers.file_utils import ModelOutput
from transformers import AutoConfig, AutoModel, AutoTokenizer
from transformers.tokenization_utils_fast import PreTrainedTokenizerFast
from transformers.tokenization_utils import BatchEncoding

from spacy.tokens import Doc
from thinc.api import Model, get_torch_default_device, xp2torch
from thinc.types import ArgsKwargs

import logging

from ..data_classes import FullTransformerBatch, WordpieceBatch, HFObjects
from ..util import maybe_flush_pytorch_cache
from ..util import log_gpu_memory, log_batch_size
from ..layers._util import replace_listener, replace_listener_cfg
from ..truncate import truncate_oversize_splits
from ..align import get_alignment, get_alignment_via_offset_mapping
from .hf_wrapper import HFWrapper


class TransformerModel(Model):
    def __init__(
        self,
        name: str,
        get_spans: Callable,
        tokenizer_config: dict = {},
        transformer_config: dict = {},
        mixed_precision: bool = False,
        grad_scaler_config: dict = {},
    ):
        """
        get_spans (Callable[[List[Doc]], List[Span]]):
            A function to extract spans from the batch of Doc objects.
            This is used to manage long documents, by cutting them into smaller
            sequences before running the transformer. The spans are allowed to
            overlap, and you can also omit sections of the Doc if they are not
            relevant.
        tokenizer_config (dict): Settings to pass to the transformers tokenizer.
        transformer_config (dict): Settings to pass to the transformers forward pass.
        """
        hf_model = HFObjects(None, None, None, tokenizer_config, transformer_config)
        wrapper = HFWrapper(
            hf_model,
            convert_inputs=_convert_transformer_inputs,
            convert_outputs=_convert_transformer_outputs,
            mixed_precision=mixed_precision,
            grad_scaler_config=grad_scaler_config,
        )
        super().__init__(
            "transformer",
            forward,
            init=init,
            layers=[wrapper],
            dims={"nO": None},
            attrs={
                "get_spans": get_spans,
                "name": name,
                "set_transformer": set_pytorch_transformer,
                "has_transformer": False,
                "flush_cache_chance": 0.0,
                "replace_listener": replace_listener,
                "replace_listener_cfg": replace_listener_cfg,
            },
        )

    @property
    def tokenizer(self):
        return self.layers[0].shims[0]._hfmodel.tokenizer

    @property
    def transformer(self):
        return self.layers[0].shims[0]._hfmodel.transformer

    @property
    def _init_tokenizer_config(self):
        return self.layers[0].shims[0]._hfmodel._init_tokenizer_config

    @property
    def _init_transformer_config(self):
        return self.layers[0].shims[0]._hfmodel._init_transformer_config

    def copy(self):
        """
        Create a copy of the model, its attributes, and its parameters. Any child
        layers will also be deep-copied. The copy will receive a distinct `model.id`
        value.
        """
        copied = TransformerModel(self.name, self.attrs["get_spans"])
        params = {}
        for name in self.param_names:
            params[name] = self.get_param(name) if self.has_param(name) else None
        copied.params = copy.deepcopy(params)
        copied.dims = copy.deepcopy(self._dims)
        copied.layers[0] = copy.deepcopy(self.layers[0])
        for name in self.grad_names:
            copied.set_grad(name, self.get_grad(name).copy())
        return copied


def set_logger(model, out_file):
    """Add a logger that will log memory usage to the given file.

    Used to debug OOM errors.
    """
    logging.basicConfig(
        level="INFO", format="%(asctime)s:%(levelname)s: %(message)s", stream=out_file
    )
    model.attrs["logger"] = logging.getLogger(__name__)


def set_pytorch_transformer(model, hf_model: HFObjects):
    if model.attrs["has_transformer"]:
        raise ValueError("Cannot set second transformer.")
    model.layers[0].shims[0]._model = hf_model.transformer
    model.layers[0].shims[0]._hfmodel.tokenizer = hf_model.tokenizer
    model.layers[0].shims[0]._hfmodel.transformer = hf_model.transformer
    model.layers[0].shims[0]._hfmodel.vocab_file_contents = hf_model.vocab_file_contents
    model.attrs["has_transformer"] = True
    model.set_dim("nO", hf_model.transformer.config.hidden_size)


def init(model: TransformerModel, X=None, Y=None):
    if model.attrs["has_transformer"]:
        return
    name = model.attrs["name"]
    tok_cfg = model._init_tokenizer_config
    trf_cfg = model._init_transformer_config
    hf_model = huggingface_from_pretrained(name, tok_cfg, trf_cfg)
    model.attrs["set_transformer"](model, hf_model)
    tokenizer = model.tokenizer
    # Call the model with a batch of inputs to infer the width
    if X:
        # If we're dealing with actual texts, do the work to setup the wordpieces
        # batch properly
        docs = X
        get_spans = model.attrs["get_spans"]
        nested_spans = get_spans(docs)
        flat_spans = []
        for doc_spans in nested_spans:
            flat_spans.extend(doc_spans)
        token_data = huggingface_tokenize(tokenizer, [span.text for span in flat_spans])
        wordpieces = WordpieceBatch.from_batch_encoding(token_data)
        if "offset_mapping" in token_data:
            align = get_alignment_via_offset_mapping(
                flat_spans,
                token_data["offset_mapping"],
            )
        else:
            align = get_alignment(
                flat_spans, wordpieces.strings, tokenizer.all_special_tokens
            )
        wordpieces, align = truncate_oversize_splits(
            wordpieces, align, tokenizer.model_max_length
        )
    else:
        texts = ["hello world", "foo bar"]
        token_data = huggingface_tokenize(tokenizer, texts)
        wordpieces = WordpieceBatch.from_batch_encoding(token_data)
    model.layers[0].initialize(X=wordpieces)
    model_output = model.layers[0].predict(wordpieces)
    model.set_dim("nO", model_output.last_hidden_state.shape[-1])


def forward(
    model: TransformerModel, docs: List[Doc], is_train: bool
) -> Tuple[FullTransformerBatch, Callable]:
    tokenizer = model.tokenizer
    get_spans = model.attrs["get_spans"]
    transformer = model.layers[0]

    nested_spans = get_spans(docs)
    flat_spans = []
    for doc_spans in nested_spans:
        flat_spans.extend(doc_spans)
    # Flush the PyTorch cache every so often. It seems to help with memory :(
    # This shouldn't be necessary, I'm not sure what I'm doing wrong?
    maybe_flush_pytorch_cache(chance=model.attrs.get("flush_cache_chance", 0))
    if "logger" in model.attrs:
        log_gpu_memory(model.attrs["logger"], "begin forward")
    batch_encoding = huggingface_tokenize(tokenizer, [span.text for span in flat_spans])
    wordpieces = WordpieceBatch.from_batch_encoding(batch_encoding)
    if "logger" in model.attrs:
        log_batch_size(model.attrs["logger"], wordpieces, is_train)
    if "offset_mapping" in batch_encoding:
        align = get_alignment_via_offset_mapping(
            flat_spans,
            batch_encoding["offset_mapping"],
        )
    else:
        align = get_alignment(
            flat_spans, wordpieces.strings, tokenizer.all_special_tokens
        )
    wordpieces, align = truncate_oversize_splits(
        wordpieces, align, tokenizer.model_max_length
    )
    model_output, bp_tensors = transformer(wordpieces, is_train)
    if "logger" in model.attrs:
        log_gpu_memory(model.attrs["logger"], "after forward")
    output = FullTransformerBatch(
        spans=nested_spans,
        wordpieces=wordpieces,
        model_output=model_output,
        align=align,
    )
    if "logger" in model.attrs:
        log_gpu_memory(model.attrs["logger"], "return from forward")

    def backprop_transformer(d_output: FullTransformerBatch) -> List[Doc]:
        if "logger" in model.attrs:
            log_gpu_memory(model.attrs["logger"], "Begin backprop")
        _ = bp_tensors(d_output.model_output)
        if "logger" in model.attrs:
            log_gpu_memory(model.attrs["logger"], "After backprop")
        return docs

    return output, backprop_transformer


def _convert_transformer_inputs(model, wps: WordpieceBatch, is_train):
    # Adapter for the HFWrapper. See https://thinc.ai/docs/usage-frameworks

    hf_device = model.shims[0]._hfmodel.transformer.device
    kwargs = {
        "input_ids": xp2torch(wps.input_ids, device=hf_device),
        "attention_mask": xp2torch(wps.attention_mask, device=hf_device),
    }
    if wps.token_type_ids is not None:
        kwargs["token_type_ids"] = xp2torch(wps.token_type_ids, device=hf_device)
    return ArgsKwargs(args=(), kwargs=kwargs), lambda dX: []


def _convert_transformer_outputs(model, inputs_outputs, is_train):
    _, model_output = inputs_outputs

    def backprop(d_model_output: ModelOutput) -> ArgsKwargs:
        return ArgsKwargs(
            args=(model_output.last_hidden_state,),
            kwargs={"grad_tensors": d_model_output.values()},
        )

    return model_output, backprop


def huggingface_from_pretrained(
    source: Union[Path, str],
    tok_config: Dict,
    trf_config: Dict,
    config_cls=AutoConfig,
    model_cls=AutoModel,
    tokenizer_cls=AutoTokenizer,
) -> HFObjects:
    """Create a Huggingface transformer model from pretrained weights. Will
    download the model if it is not already downloaded.

    source (Union[str, Path]): The name of the model or a path to it, such as
        'bert-base-cased'.
    tok_config (dict): Settings to pass to the tokenizer.
    trf_config (dict): Settings to pass to the transformer.
    """
    if isinstance(source, Path):
        str_path = str(source.absolute())
    else:
        str_path = source
    tokenizer = tokenizer_cls.from_pretrained(str_path, **tok_config)
    vocab_file_contents = None
    if hasattr(tokenizer, "vocab_file"):
        with open(tokenizer.vocab_file, "rb") as fileh:
            vocab_file_contents = fileh.read()
    trf_config["return_dict"] = True
    config = config_cls.from_pretrained(str_path, **trf_config)
    transformer = model_cls.from_pretrained(str_path, config=config)
    torch_device = get_torch_default_device()
    transformer.to(torch_device)
    return HFObjects(tokenizer, transformer, vocab_file_contents)


def huggingface_tokenize(tokenizer, texts: List[str]) -> BatchEncoding:
    """Apply a Huggingface tokenizer to a batch of texts."""

    # Use NumPy arrays rather than PyTorch tensors to avoid a lot of
    # host <-> device transfers during tokenization and post-processing
    # when a GPU is used.
    token_data = tokenizer(
        texts,
        add_special_tokens=True,
        return_attention_mask=True,
        return_offsets_mapping=isinstance(tokenizer, PreTrainedTokenizerFast),
        return_tensors="np",
        return_token_type_ids=None,  # Sets to model default
        padding="longest",
    )
    token_data["input_texts"] = []
    for i in range(len(token_data["input_ids"])):
        wp_texts = tokenizer.convert_ids_to_tokens(token_data["input_ids"][i])
        token_data["input_texts"].append(wp_texts)
    token_data["pad_token"] = tokenizer.pad_token
    return token_data



from typing import Callable, List, Optional, Tuple, cast
import numpy
from spacy.util import all_equal
from transformers.file_utils import ModelOutput
from transformers.modeling_outputs import BaseModelOutput
from thinc.api import Model
from thinc.types import Ragged, Floats2d
from ..data_classes import TransformerData
from ..align import apply_alignment


def trfs2arrays(
    pooling: Model[Ragged, Floats2d], grad_factor: float
) -> Model[List[TransformerData], List[Floats2d]]:
    """Pool transformer data into token-aligned tensors."""
    return Model(
        "trfs2arrays",
        forward,
        layers=[pooling],
        attrs={"grad_factor": grad_factor},
    )


def forward(model: Model, trf_datas: List[TransformerData], is_train: bool):
    pooling: Model[Ragged, Floats2d] = model.layers[0]
    grad_factor = model.attrs["grad_factor"]
    zero_outputs: List[Tuple[int, Floats2d]] = []
    backprops_alignment: List[Optional[Callable]] = []
    aligned_outputs: List[Tuple[int, Ragged]] = []

    # For zero-length documents, we could cache the output width by iterating
    # through the batch outputs and retrieving the shape of a non-zero length
    # Doc. This, however, is not fool-proof as one can pass an entire batch of
    # zero-length Docs to the transformer model (at least during prediction).
    # Instead of being conditionally correct, we'll explicitly leave the width as
    # zero in these cases as the effective length of the resultant tensor is zero anyway.
    output_width = 0

    for i, trf_data in enumerate(trf_datas):
        if not isinstance(trf_data, TransformerData):
            raise ValueError(
                "Expected spacy_transformers.data_classes.TransformerData "
                f"in trf_data, got: {type(trf_data)}\n"
                "Check that your pipeline contains a transformer component "
                "with a spacy-transformers TransformerModel architecture."
            )
        if "last_hidden_state" in trf_data.model_output:
            tensor_t_i = cast(BaseModelOutput, trf_data.model_output).last_hidden_state
            if tensor_t_i.size == 0:
                # This can happen during prediction/initialization if the transformer pipe was disabled/not executed and one of the inputs
                # was of length zero. This causes the listenener to generate a zero-sized (in the sequence length dim) TransformerData
                # output and pass it downstream.
                zero_outputs.append((i, model.ops.alloc2f(0, output_width)))
                backprops_alignment.append(None)
            else:
                # This is the general case for non-zero length documents.
                src = model.ops.reshape2f(tensor_t_i, -1, trf_data.width)  # type: ignore
                dst, get_d_src = apply_alignment(model.ops, trf_data.align, src)
                aligned_outputs.append((i, dst))
                backprops_alignment.append(get_d_src)
        else:
            # This can happen during prediction/training for zero-length documents. Since zero-length docs
            # are implicitly ignored in the span generation stage, the transformer model does not return any
            # predictions for them and subsequently, FullTransformerBatch.split_by_doc() generates an empty
            # TransformerData.
            zero_outputs.append((i, model.ops.alloc2f(0, output_width)))
            backprops_alignment.append(None)

    pooling_outputs, backprop_pooling = concat_pooling_forward(
        pooling, [dst for _, dst in aligned_outputs], is_train
    )

    # Interleave the zero and non-zero outputs into the final result.
    outputs: List[Optional[Floats2d]] = [None] * (
        len(zero_outputs) + len(aligned_outputs)
    )
    for i, zero_output in zero_outputs:
        outputs[i] = zero_output
    for (i, _), pooling_output in zip(aligned_outputs, pooling_outputs):
        outputs[i] = pooling_output

    def backprop_trf_to_tensor(d_outputs: List[Floats2d]) -> List[TransformerData]:
        d_trf_datas: List[TransformerData] = []

        # Only update the gradients that are relevant for pooling.
        d_pooling = backprop_pooling([d_outputs[i] for i, _ in aligned_outputs])
        for (i, _), d_pooling_i in zip(aligned_outputs, d_pooling):
            d_outputs[i] = d_pooling_i

        to_zip = (trf_datas, d_outputs, backprops_alignment)
        assert all_equal(len(x) for x in to_zip)  # type: ignore
        zipped = zip(*to_zip)
        for trf_data, d_output, get_d_src in zipped:
            if "last_hidden_state" not in trf_data.model_output:
                # This gradient belongs to a zero-length doc and must be ignored as it doesn't have a corresponding
                # output from the transformer model (due to empty documents being skipped during the span generation
                # stage in the forward pass).
                assert len(d_output) == 0
                assert get_d_src is None
                continue

            assert get_d_src is not None
            d_model_output = ModelOutput(
                last_hidden_state=model.ops.alloc(
                    trf_data.model_output.last_hidden_state.shape,  # type: ignore
                    dtype=trf_data.model_output.last_hidden_state.dtype,  # type: ignore
                )
            )
            d_src = get_d_src(d_output)
            d_src *= grad_factor
            d_model_output["last_hidden_state"] = d_src.reshape(
                cast(BaseModelOutput, trf_data.model_output).last_hidden_state.shape
            )
            d_trf_datas.append(
                TransformerData(
                    model_output=d_model_output,
                    wordpieces=trf_data.wordpieces,
                    align=trf_data.align,
                )
            )
        return d_trf_datas

    assert len(outputs) == len(trf_datas)
    return outputs, backprop_trf_to_tensor


def concat_pooling_forward(
    pooling: Model[Ragged, Floats2d], X: List[Ragged], is_train: bool
):
    xp = pooling.ops.xp

    datas = []
    lens = []
    doc_lens = []
    for X_doc_data in X:
        datas.append(X_doc_data.dataXd)
        lens.append(X_doc_data.lengths)
        doc_lens.append(len(X_doc_data.lengths))

    X_flat = Ragged(xp.concatenate(datas, axis=0), xp.concatenate(lens, axis=0))
    Y_pooled, pooling_backprop = pooling(X_flat, is_train)
    Y = xp.split(Y_pooled, numpy.cumsum(doc_lens)[:-1])

    def backprop(dY):
        dY_pooled_flat = xp.concatenate(dY)
        dY_flat = pooling_backprop(dY_pooled_flat).dataXd

        dY = []
        for X_doc_data in X:
            doc_unpooled_len = X_doc_data.dataXd.shape[0]
            dY.append(Ragged(dY_flat[:doc_unpooled_len], X_doc_data.lengths))
            dY_flat = dY_flat[doc_unpooled_len:]

        return dY

    return Y, backprop
